# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pickle

import pyomo.common.unittest as unittest
from pyomo.core.expr.numvalue import (
    NumericValue,
    is_fixed,
    is_constant,
    is_potentially_variable,
    value,
)
import pyomo.kernel
from pyomo.core.tests.unit.kernel.test_dict_container import (
    _TestActiveDictContainerBase,
)
from pyomo.core.tests.unit.kernel.test_tuple_container import (
    _TestActiveTupleContainerBase,
)
from pyomo.core.tests.unit.kernel.test_list_container import (
    _TestActiveListContainerBase,
)
from pyomo.core.kernel.base import ICategorizedObject
from pyomo.core.kernel.expression import (
    IIdentityExpression,
    noclone,
    IExpression,
    expression,
    data_expression,
    expression_dict,
    expression_tuple,
    expression_list,
)
from pyomo.core.kernel.variable import variable
from pyomo.core.kernel.parameter import parameter
from pyomo.core.kernel.objective import objective
from pyomo.core.kernel.block import block

try:
    import numpy

    numpy_available = True
except:
    numpy_available = False


class Test_noclone(unittest.TestCase):
    def test_is_named_expression_type(self):
        e = expression()
        self.assertEqual(e.is_named_expression_type(), True)

    def test_arg(self):
        e = expression()
        self.assertEqual(e.arg(0), None)
        e.expr = 1
        self.assertEqual(e.arg(0), 1)
        with self.assertRaises(KeyError):
            e.arg(1)

    def test_init_non_NumericValue(self):
        types = [None, 1, 1.1, True, ""]
        if numpy_available:
            types.extend([numpy.float32(1), numpy.bool_(True), numpy.int32(1)])
        types.append(block())
        types.append(block)
        for obj in types:
            self.assertEqual(noclone(obj), obj)
            self.assertIs(type(noclone(obj)), type(obj))

    def test_init_NumericValue(self):
        v = variable()
        p = parameter()
        e = expression()
        d = data_expression()
        o = objective()
        for obj in (
            v,
            v + 1,
            v**2,
            p,
            p + 1,
            p**2,
            e,
            e + 1,
            e**2,
            d,
            d + 1,
            d**2,
            o,
            o + 1,
            o**2,
        ):
            self.assertTrue(isinstance(noclone(obj), NumericValue))
            self.assertTrue(isinstance(noclone(obj), IIdentityExpression))
            self.assertIs(noclone(obj).expr, obj)

    def test_pprint(self):
        import pyomo.kernel

        # Not really testing what the output is, just that
        # an error does not occur. The pprint functionality
        # is still in the early stages.
        v = variable()
        e = noclone(v**2)
        pyomo.kernel.pprint(e)
        pyomo.kernel.pprint(e, indent=1)
        b = block()
        b.e = expression(expr=e)
        pyomo.kernel.pprint(e)
        pyomo.kernel.pprint(b)
        m = block()
        m.b = b
        pyomo.kernel.pprint(e)
        pyomo.kernel.pprint(b)
        pyomo.kernel.pprint(m)
        # tests compatibility with _ToStringVisitor
        pyomo.kernel.pprint(noclone(v) + 1)
        pyomo.kernel.pprint(noclone(v + 1))
        x = variable()
        y = variable()
        pyomo.kernel.pprint(y + x * noclone(noclone(x * y)))
        pyomo.kernel.pprint(y + noclone(noclone(x * y)) * x)

    def test_pickle(self):
        v = variable()
        e = noclone(v)
        self.assertEqual(type(e), expression)
        self.assertIs(type(e.expr), variable)
        eup = pickle.loads(pickle.dumps(e))
        self.assertEqual(type(eup), expression)
        self.assertTrue(e is not eup)
        self.assertIs(type(eup.expr), variable)
        self.assertIs(type(e.expr), variable)
        self.assertTrue(eup.expr is not e.expr)

        del e
        del v

        v = variable(value=1)
        b = block()
        b.v = v
        eraw = b.v + 1
        b.e = 1 + noclone(eraw)
        bup = pickle.loads(pickle.dumps(b))
        self.assertTrue(isinstance(bup.e, NumericValue))
        self.assertEqual(value(bup.e), 3.0)
        b.v.value = 2
        self.assertEqual(value(b.e), 4.0)
        self.assertEqual(value(bup.e), 3.0)
        bup.v.value = -1
        self.assertEqual(value(b.e), 4.0)
        self.assertEqual(value(bup.e), 1.0)

        self.assertIs(b.v.parent, b)
        self.assertIs(bup.v.parent, bup)

        del b.v

    def test_call(self):
        e = noclone(None)
        self.assertIs(e, None)
        e = noclone(1)
        self.assertEqual(e, 1)
        p = parameter()
        p.value = 2
        e = noclone(p + 1)
        self.assertEqual(e(), 3)

    def test_is_constant(self):
        v = variable()
        e = noclone(v)
        self.assertEqual(e.is_constant(), False)
        self.assertEqual(is_constant(e), False)
        v.fix(1)
        self.assertEqual(e.is_constant(), False)
        self.assertEqual(is_constant(e), False)

        p = parameter()
        e = noclone(p)
        self.assertEqual(p.is_constant(), False)
        self.assertEqual(is_constant(p), False)

        self.assertEqual(is_constant(noclone(1)), True)

    def test_is_fixed(self):
        v = variable()
        e = noclone(v + 1)
        self.assertEqual(e.is_fixed(), False)
        self.assertEqual(is_fixed(e), False)
        v.fix()
        self.assertEqual(e.is_fixed(), True)
        self.assertEqual(is_fixed(e), True)

        e = noclone(parameter())
        self.assertEqual(e.is_fixed(), True)
        self.assertEqual(is_fixed(e), True)

    def testis_potentially_variable(self):
        e = noclone(variable())
        self.assertEqual(e.is_potentially_variable(), True)
        self.assertEqual(is_potentially_variable(e), True)
        e = noclone(parameter())
        self.assertEqual(e.is_potentially_variable(), False)
        self.assertEqual(is_potentially_variable(e), False)
        e = noclone(expression())
        self.assertEqual(e.is_potentially_variable(), True)
        self.assertEqual(is_potentially_variable(e), True)
        e = noclone(data_expression())
        self.assertEqual(e.is_potentially_variable(), False)
        self.assertEqual(is_potentially_variable(e), False)

    def test_polynomial_degree(self):
        e = noclone(parameter())
        self.assertEqual(e.polynomial_degree(), 0)
        e = noclone(parameter(value=1))
        self.assertEqual(e.polynomial_degree(), 0)
        v = variable()
        v.value = 2
        e = noclone(v + 1)
        self.assertEqual(e.polynomial_degree(), 1)
        e = noclone(v**2 + v + 1)
        self.assertEqual(e.polynomial_degree(), 2)
        v.fix()
        self.assertEqual(e.polynomial_degree(), 0)
        e = noclone(v**v)
        self.assertEqual(e.polynomial_degree(), 0)
        v.free()
        self.assertEqual(e.polynomial_degree(), None)

    def test_is_expression_type(self):
        for obj in (
            variable(),
            parameter(),
            objective(),
            expression(),
            data_expression(),
        ):
            self.assertEqual(noclone(obj).is_expression_type(), True)

    def test_is_parameter_type(self):
        for obj in (
            variable(),
            parameter(),
            objective(),
            expression(),
            data_expression(),
        ):
            self.assertEqual(noclone(obj).is_parameter_type(), False)

    def test_args(self):
        e = noclone(parameter() + 1)
        self.assertEqual(e.nargs(), 1)
        self.assertTrue(e.arg(0) is e.expr)

    def test_arguments(self):
        e = noclone(parameter() + 1)
        self.assertEqual(len(tuple(e.args)), 1)
        self.assertTrue(tuple(e.args)[0] is e.expr)

    def test_clone(self):
        p = parameter()
        e = noclone(p)
        self.assertTrue(e.clone() is e)
        self.assertTrue(e.clone().expr is p)
        sube = p**2 + 1
        e = noclone(sube)
        self.assertTrue(e.clone() is e)
        self.assertTrue(e.clone().expr is sube)

    def test_division_behavior(self):
        # make sure integers involved in Pyomo expression
        # use __future__ behavior
        e = noclone(parameter(value=2))
        self.assertIs(type(e.expr), parameter)
        self.assertEqual((1 / e)(), 0.5)
        self.assertEqual((parameter(1) / e)(), 0.5)
        self.assertEqual((1 / e.expr()), 0.5)

    def test_to_string(self):
        b = block()
        p = parameter()
        e = noclone(p**2)
        self.assertEqual(str(e.expr), "<parameter>**2")
        self.assertEqual(str(e), "<data_expression>")
        self.assertEqual(e.to_string(), "(<parameter>**2)")
        self.assertEqual(e.to_string(verbose=False), "(<parameter>**2)")
        self.assertEqual(
            e.to_string(verbose=True), "<data_expression>{pow(<parameter>, 2)}"
        )
        b.e = e
        b.p = p
        self.assertNotEqual(p.name, None)
        self.assertEqual(e.to_string(verbose=True), "e{pow(" + p.name + ", 2)}")
        self.assertEqual(e.to_string(verbose=True), "e{pow(p, 2)}")
        del b.e
        del b.p


class _Test_expression_base:
    _ctype_factory = None

    def test_pprint(self):
        import pyomo.kernel

        # Not really testing what the output is, just that
        # an error does not occur. The pprint functionality
        # is still in the early stages.
        p = parameter()
        e = self._ctype_factory(p**2)
        pyomo.kernel.pprint(e)
        b = block()
        b.e = e
        pyomo.kernel.pprint(e)
        pyomo.kernel.pprint(b)
        m = block()
        m.b = b
        pyomo.kernel.pprint(e)
        pyomo.kernel.pprint(b)
        pyomo.kernel.pprint(m)

    def test_pickle(self):
        e = self._ctype_factory(expr=1.0)
        self.assertEqual(type(e.expr), float)
        self.assertEqual(e.expr, 1.0)
        self.assertEqual(e.parent, None)
        eup = pickle.loads(pickle.dumps(e))
        self.assertEqual(type(eup.expr), float)
        self.assertEqual(eup.expr, 1.0)
        self.assertEqual(eup.parent, None)
        b = block()
        b.e = e
        self.assertIs(e.parent, b)
        bup = pickle.loads(pickle.dumps(b))
        eup = bup.e
        self.assertEqual(type(eup.expr), float)
        self.assertEqual(eup.expr, 1.0)
        self.assertIs(eup.parent, bup)

    def test_init_no_args(self):
        e = self._ctype_factory()
        self.assertTrue(e.parent is None)
        self.assertEqual(e.ctype, IExpression)
        self.assertTrue(e.expr is None)

    def test_init_args(self):
        e = self._ctype_factory(1.0)
        self.assertTrue(e.parent is None)
        self.assertEqual(e.ctype, IExpression)
        self.assertTrue(e.expr is not None)

    def test_type(self):
        e = self._ctype_factory()
        self.assertTrue(isinstance(e, ICategorizedObject))
        self.assertTrue(isinstance(e, IExpression))
        self.assertTrue(isinstance(e, NumericValue))
        self.assertTrue(isinstance(e, IIdentityExpression))

    def test_call(self):
        e = self._ctype_factory()
        self.assertEqual(e(), None)
        e.expr = 1
        self.assertEqual(e(), 1)
        p = parameter()
        p.value = 2
        e.expr = p + 1
        self.assertEqual(e(), 3)

    def test_is_constant(self):
        e = self._ctype_factory()
        self.assertEqual(e.is_constant(), False)
        self.assertEqual(is_constant(e), False)
        e.expr = 1
        self.assertEqual(e.is_constant(), False)
        self.assertEqual(is_constant(e), False)
        p = parameter()
        self.assertEqual(p.is_constant(), False)
        self.assertEqual(is_constant(p), False)
        p.value = 2
        e.expr = p + 1
        self.assertEqual(e.is_constant(), False)
        self.assertEqual(is_constant(e), False)

    def test_is_expression_type(self):
        e = self._ctype_factory()
        self.assertEqual(e.is_expression_type(), True)

    def test_is_parameter_type(self):
        e = self._ctype_factory()
        self.assertEqual(e.is_parameter_type(), False)

    def test_args(self):
        e = self._ctype_factory()
        p = parameter()
        e.expr = p + 1
        self.assertEqual(e.nargs(), 1)
        self.assertTrue(e.arg(0) is e.expr)

    def test_arguments(self):
        e = self._ctype_factory()
        p = parameter()
        e.expr = p + 1
        self.assertEqual(len(tuple(e.args)), 1)
        self.assertTrue(tuple(e.args)[0] is e.expr)

    def test_clone(self):
        e = self._ctype_factory()
        self.assertTrue(e.clone() is e)
        p = parameter()
        e.expr = p
        self.assertTrue(e.clone() is e)
        self.assertTrue(e.clone().expr is p)
        sube = p**2 + 1
        e.expr = sube
        self.assertTrue(e.clone() is e)
        self.assertTrue(e.clone().expr is sube)

    def test_division_behavior(self):
        # make sure integers involved in Pyomo expression
        # use __future__ behavior
        e = self._ctype_factory()
        e.expr = 2
        self.assertIs(type(e.expr), int)
        self.assertEqual((1 / e)(), 0.5)
        self.assertEqual((parameter(1) / e)(), 0.5)
        self.assertEqual((1 / e.expr), 0.5)

    def test_to_string(self):
        b = block()
        e = self._ctype_factory()
        label = str(e)
        self.assertNotEqual(label, None)
        self.assertEqual(e.name, None)

        self.assertEqual(str(e.expr), "None")
        self.assertEqual(str(e), label)
        self.assertEqual(e.to_string(), label + "{Undefined}")
        self.assertEqual(e.to_string(verbose=False), label + "{Undefined}")
        self.assertEqual(e.to_string(verbose=True), label + "{Undefined}")
        b.e = e
        self.assertNotEqual(e.name, None)
        self.assertEqual(e.to_string(verbose=True), "e{Undefined}")
        del b.e
        self.assertEqual(e.name, None)

        e.expr = 1
        self.assertEqual(str(e.expr), "1")
        self.assertEqual(str(e), label)
        self.assertEqual(e.to_string(), "1")
        self.assertEqual(e.to_string(verbose=False), "1")
        self.assertEqual(e.to_string(verbose=True), label + "{1}")
        b.e = e
        self.assertNotEqual(e.name, None)
        self.assertEqual(e.to_string(verbose=True), "e{1}")
        del b.e
        self.assertEqual(e.name, None)

        p = parameter()
        e.expr = p**2
        self.assertEqual(str(e.expr), "<parameter>**2")
        self.assertEqual(str(e), label)
        self.assertEqual(e.to_string(), "(<parameter>**2)")
        self.assertEqual(e.to_string(verbose=False), "(<parameter>**2)")
        self.assertEqual(e.to_string(verbose=True), label + "{pow(<parameter>, 2)}")
        b.e = e
        b.p = p
        self.assertNotEqual(e.name, None)
        self.assertNotEqual(p.name, None)
        self.assertEqual(e.to_string(verbose=True), e.name + "{pow(" + p.name + ", 2)}")
        self.assertEqual(e.to_string(verbose=True), "e{pow(p, 2)}")
        del b.e
        del b.p

    def test_iadd(self):
        # make sure simple for loops that look like they
        # create a new expression do not modify the named
        # expression
        e = self._ctype_factory(1.0)
        expr = 0.0
        for v in [1.0, e]:
            expr += v
        self.assertEqual(e.expr, 1)
        self.assertEqual(expr(), 2)
        expr = 0.0
        for v in [e, 1.0]:
            expr += v
        self.assertEqual(e.expr, 1)
        self.assertEqual(expr(), 2)

    def test_isub(self):
        # make sure simple for loops that look like they
        # create a new expression do not modify the named
        # expression
        e = self._ctype_factory(1.0)
        expr = 0.0
        for v in [1.0, e]:
            expr -= v
        self.assertEqual(e.expr, 1)
        self.assertEqual(expr(), -2)
        expr = 0.0
        for v in [e, 1.0]:
            expr -= v
        self.assertEqual(e.expr, 1)
        self.assertEqual(expr(), -2)

    def test_imul(self):
        # make sure simple for loops that look like they
        # create a new expression do not modify the named
        # expression
        e = self._ctype_factory(3.0)
        expr = 1.0
        for v in [2.0, e]:
            expr *= v
        self.assertEqual(e.expr, 3)
        self.assertEqual(expr(), 6)
        expr = 1.0
        for v in [e, 2.0]:
            expr *= v
        self.assertEqual(e.expr, 3)
        self.assertEqual(expr(), 6)

    def test_idiv(self):
        # make sure simple for loops that look like they
        # create a new expression do not modify the named
        # expression
        # floating point division
        e = self._ctype_factory(3.0)
        expr = e
        for v in [2.0, 1.0]:
            expr /= v
        self.assertEqual(e.expr, 3)
        self.assertEqual(expr(), 1.5)
        expr = e
        for v in [1.0, 2.0]:
            expr /= v
        self.assertEqual(e.expr, 3)
        self.assertEqual(expr(), 1.5)
        # note that integer division does not occur within
        # Pyomo expressions
        e = self._ctype_factory(3)
        expr = e
        for v in [2, 1]:
            expr /= v
        self.assertEqual(e.expr, 3)
        self.assertEqual(expr(), 1.5)
        expr = e
        for v in [1, 2]:
            expr /= v
        self.assertEqual(e.expr, 3)
        self.assertEqual(expr(), 1.5)

    def test_ipow(self):
        # make sure simple for loops that look like they
        # create a new expression do not modify the named
        # expression
        e = self._ctype_factory(3.0)
        expr = e
        for v in [2.0, 1.0]:
            expr **= v
        self.assertEqual(e.expr, 3)
        self.assertEqual(expr(), 9)
        expr = e
        for v in [1.0, 2.0]:
            expr **= v
        self.assertEqual(e.expr, 3)
        self.assertEqual(expr(), 9)


class Test_expression(_Test_expression_base, unittest.TestCase):
    _ctype_factory = expression

    def test_associativity(self):
        x = variable()
        y = variable()
        pyomo.kernel.pprint(y + x * expression(expression(x * y)))
        pyomo.kernel.pprint(y + expression(expression(x * y)) * x)

    def test_ctype(self):
        e = expression()
        self.assertIs(e.ctype, IExpression)
        self.assertIs(type(e), expression)
        self.assertIs(type(e)._ctype, IExpression)

    def test_is_fixed(self):
        e = self._ctype_factory()
        self.assertEqual(e.is_fixed(), True)
        self.assertEqual(is_fixed(e), True)
        e.expr = 1
        self.assertEqual(e.is_fixed(), True)
        self.assertEqual(is_fixed(e), True)
        v = variable()
        v.value = 2
        e.expr = v + 1
        self.assertEqual(e.is_fixed(), False)
        self.assertEqual(is_fixed(e), False)
        v.fix()
        self.assertEqual(e.is_fixed(), True)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(e(), 3)

    def testis_potentially_variable(self):
        e = self._ctype_factory()
        self.assertEqual(e.is_potentially_variable(), True)
        self.assertEqual(is_potentially_variable(e), True)
        e.expr = 1
        self.assertEqual(e.is_potentially_variable(), True)
        self.assertEqual(is_potentially_variable(e), True)
        v = variable()
        v.value = 2
        e.expr = v + 1
        self.assertEqual(e.is_potentially_variable(), True)
        self.assertEqual(is_potentially_variable(e), True)
        v.fix()
        e.expr = v + 1
        self.assertEqual(e.is_potentially_variable(), True)
        self.assertEqual(is_potentially_variable(e), True)
        self.assertEqual(e(), 3)

    def test_polynomial_degree(self):
        e = self._ctype_factory()
        e.expr = 1
        self.assertEqual(e.polynomial_degree(), 0)
        v = variable()
        v.value = 2
        e.expr = v + 1
        self.assertEqual(e.polynomial_degree(), 1)
        e.expr = v**2 + v + 1
        self.assertEqual(e.polynomial_degree(), 2)
        v.fix()
        self.assertEqual(e.polynomial_degree(), 0)
        e.expr = v**v
        self.assertEqual(e.polynomial_degree(), 0)
        v.free()
        self.assertEqual(e.polynomial_degree(), None)


class Test_data_expression(_Test_expression_base, unittest.TestCase):
    _ctype_factory = data_expression

    def test_associativity(self):
        x = parameter()
        y = parameter()
        pyomo.kernel.pprint(y + x * data_expression(data_expression(x * y)))
        pyomo.kernel.pprint(y + data_expression(data_expression(x * y)) * x)

    def test_ctype(self):
        e = data_expression()
        self.assertIs(e.ctype, IExpression)
        self.assertIs(type(e), data_expression)
        self.assertIs(type(e)._ctype, IExpression)

    def test_bad_init(self):
        e = self._ctype_factory(expr=1.0)
        self.assertEqual(e.expr, 1.0)

        v = variable()
        with self.assertRaises(ValueError):
            e = self._ctype_factory(expr=v)

    def test_bad_assignment(self):
        e = self._ctype_factory(expr=1.0)
        self.assertEqual(e.expr, 1.0)

        v = variable()
        with self.assertRaises(ValueError):
            e.expr = v + 1

    def test_is_fixed(self):
        e = self._ctype_factory()
        self.assertEqual(e.is_fixed(), True)
        self.assertEqual(is_fixed(e), True)
        e.expr = 1
        self.assertEqual(e.is_fixed(), True)
        self.assertEqual(is_fixed(e), True)
        p = parameter()
        e.expr = p**2
        self.assertEqual(e.is_fixed(), True)
        self.assertEqual(is_fixed(e), True)
        a = self._ctype_factory()
        e.expr = (a * p) ** 2 / (p + 5)
        self.assertEqual(e.is_fixed(), True)
        self.assertEqual(is_fixed(e), True)
        a.expr = 2.0
        p.value = 5.0
        self.assertEqual(e.is_fixed(), True)
        self.assertEqual(is_fixed(e), True)
        self.assertEqual(e(), 10.0)

        v = variable()
        with self.assertRaises(ValueError):
            e.expr = v + 1

    def testis_potentially_variable(self):
        e = self._ctype_factory()
        self.assertEqual(e.is_potentially_variable(), False)
        self.assertEqual(is_potentially_variable(e), False)
        e.expr = 1
        self.assertEqual(e.is_potentially_variable(), False)
        self.assertEqual(is_potentially_variable(e), False)
        p = parameter()
        e.expr = p**2
        self.assertEqual(e.is_potentially_variable(), False)
        self.assertEqual(is_potentially_variable(e), False)
        a = self._ctype_factory()
        e.expr = (a * p) ** 2 / (p + 5)
        self.assertEqual(e.is_potentially_variable(), False)
        self.assertEqual(is_potentially_variable(e), False)
        a.expr = 2.0
        p.value = 5.0
        self.assertEqual(e.is_potentially_variable(), False)
        self.assertEqual(is_potentially_variable(e), False)
        self.assertEqual(e(), 10.0)

        v = variable()
        with self.assertRaises(ValueError):
            e.expr = v + 1

    def test_polynomial_degree(self):
        e = self._ctype_factory()
        self.assertEqual(e.polynomial_degree(), 0)
        e.expr = 1
        self.assertEqual(e.polynomial_degree(), 0)
        p = parameter()
        e.expr = p**2
        self.assertEqual(e.polynomial_degree(), 0)
        a = self._ctype_factory()
        e.expr = (a * p) ** 2 / (p + 5)
        self.assertEqual(e.polynomial_degree(), 0)
        a.expr = 2.0
        p.value = 5.0
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(e(), 10.0)

        v = variable()
        with self.assertRaises(ValueError):
            e.expr = v + 1


class Test_expression_dict(_TestActiveDictContainerBase, unittest.TestCase):
    _container_type = expression_dict
    _ctype_factory = lambda self: expression()


class Test_expression_tuple(_TestActiveTupleContainerBase, unittest.TestCase):
    _container_type = expression_tuple
    _ctype_factory = lambda self: expression()


class Test_expression_list(_TestActiveListContainerBase, unittest.TestCase):
    _container_type = expression_list
    _ctype_factory = lambda self: expression()


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.base.block import Block
from pyomo.core.base.reference import Reference
from pyomo.core.expr.visitor import identify_variables
from pyomo.common.collections import ComponentSet, ComponentMap
from pyomo.common.modeling import unique_component_name
from pyomo.util.vars_from_expressions import get_vars_from_components
from pyomo.core.base.constraint import Constraint
from pyomo.core.base.expression import Expression
from pyomo.core.base.objective import Objective
from pyomo.core.base.external import ExternalFunction
from pyomo.core.expr.visitor import StreamBasedExpressionVisitor
from pyomo.core.expr.numeric_expr import ExternalFunctionExpression
from pyomo.core.expr.numvalue import native_types, NumericValue


class _ExternalFunctionVisitor(StreamBasedExpressionVisitor):
    def __init__(self, descend_into_named_expressions=True):
        super().__init__()
        self._descend_into_named_expressions = descend_into_named_expressions
        self.named_expressions = []

    def initializeWalker(self, expr):
        self._functions = []
        self._seen = set()
        return True, None

    def beforeChild(self, parent, child, index):
        if child.__class__ in native_types:
            return False, None
        elif (
            not self._descend_into_named_expressions
            and child.is_named_expression_type()
        ):
            self.named_expressions.append(child)
            return False, None
        return True, None

    def exitNode(self, node, data):
        if type(node) is ExternalFunctionExpression:
            if id(node) not in self._seen:
                self._seen.add(id(node))
                self._functions.append(node)

    def finalizeResult(self, result):
        return self._functions


def identify_external_functions(expr):
    yield from _ExternalFunctionVisitor().walk_expression(expr)


def add_local_external_functions(block):
    ef_exprs = []
    named_expressions = []
    visitor = _ExternalFunctionVisitor(descend_into_named_expressions=False)
    for comp in block.component_data_objects(
        (Constraint, Expression, Objective), active=True
    ):
        ef_exprs.extend(visitor.walk_expression(comp.expr))
    named_expr_set = ComponentSet(visitor.named_expressions)
    # List of unique named expressions
    named_expressions = list(named_expr_set)
    while named_expressions:
        expr = named_expressions.pop()
        # Clear named expression cache so we don't re-check named expressions
        # we've seen before.
        visitor.named_expressions.clear()
        ef_exprs.extend(visitor.walk_expression(expr))
        # Only add to the stack named expressions that we have
        # not encountered yet.
        for local_expr in visitor.named_expressions:
            if local_expr not in named_expr_set:
                named_expressions.append(local_expr)
                named_expr_set.add(local_expr)

    unique_functions = []
    fcn_set = set()
    for expr in ef_exprs:
        fcn = expr._fcn
        data = (fcn._library, fcn._function)
        if data not in fcn_set:
            fcn_set.add(data)
            unique_functions.append(data)
    fcn_comp_map = {}
    for lib, name in unique_functions:
        comp_name = unique_component_name(block, "_" + name)
        comp = ExternalFunction(library=lib, function=name)
        block.add_component(comp_name, comp)
        fcn_comp_map[lib, name] = comp
    return fcn_comp_map


def create_subsystem_block(constraints, variables=None, include_fixed=False):
    """This function creates a block to serve as a subsystem with the
    specified variables and constraints. To satisfy certain writers, other
    variables that appear in the constraints must be added to the block as
    well. We call these the "input vars." They may be thought of as
    parameters in the subsystem, but we do not fix them here as it is not
    obvious that this is desired.

    Arguments
    ---------
    constraints: List
        List of Pyomo constraint data objects
    variables: List
        List of Pyomo var data objects
    include_fixed: Bool
        Indicates whether fixed variables should be attached to the block.
        This is useful if they may be unfixed at some point.

    Returns
    -------
    Block containing references to the specified constraints and variables,
    as well as other variables present in the constraints

    """
    if variables is None:
        variables = []
    block = Block(concrete=True)
    block.vars = Reference(variables)
    block.cons = Reference(constraints)
    var_set = ComponentSet(variables)
    input_vars = []
    for var in get_vars_from_components(block, Constraint, include_fixed=include_fixed):
        if var not in var_set:
            input_vars.append(var)
    block.input_vars = Reference(input_vars)
    add_local_external_functions(block)
    return block


def generate_subsystem_blocks(subsystems, include_fixed=False):
    """Generates blocks that contain subsystems of variables and constraints.

    Arguments
    ---------
    subsystems: List of tuples
        Each tuple is a list of constraints then a list of variables
        that will define a subsystem.
    include_fixed: Bool
        Indicates whether to add already fixed variables to the generated
        subsystem blocks.

    Yields
    ------
    "Subsystem blocks" containing the variables and constraints specified
    by each entry in subsystems. Variables in the constraints that are
    not specified are contained in the input_vars component.

    """
    for cons, vars in subsystems:
        block = create_subsystem_block(cons, vars, include_fixed)
        yield block, list(block.input_vars.values())


class TemporarySubsystemManager:
    """This class is a context manager for cases when we want to
    temporarily fix or deactivate certain variables or constraints
    in order to perform some solve or calculation with the resulting
    subsystem.

    """

    def __init__(
        self,
        to_fix=None,
        to_deactivate=None,
        to_reset=None,
        to_unfix=None,
        remove_bounds_on_fix=False,
    ):
        """
        Arguments
        ---------
        to_fix: List
            List of var data objects that should be temporarily fixed.
            These are restored to their original status on exit from
            this object's context manager.
        to_deactivate: List
            List of constraint data objects that should be temporarily
            deactivated. These are restored to their original status on
            exit from this object's context manager.
        to_reset: List
            List of var data objects that should be reset to their
            original values on exit from this object's context context
            manager.
        to_unfix: List
            List of var data objects to be temporarily unfixed. These are
            restored to their original status on exit from this object's
            context manager.
        remove_bounds_on_fix: Bool
            Whether bounds should be removed temporarily for fixed variables

        """
        if to_fix is None:
            to_fix = []
        if to_deactivate is None:
            to_deactivate = []
        if to_reset is None:
            to_reset = []
        if to_unfix is None:
            to_unfix = []
        if not ComponentSet(to_fix).isdisjoint(ComponentSet(to_unfix)):
            to_unfix_set = ComponentSet(to_unfix)
            both = [var for var in to_fix if var in to_unfix_set]
            var_names = "\n" + "\n".join([var.name for var in both])
            raise RuntimeError(
                f"Conflicting instructions: The following variables are present"
                " in both to_fix and to_unfix lists: {var_names}"
            )
        self._vars_to_fix = to_fix
        self._cons_to_deactivate = to_deactivate
        self._comps_to_set = to_reset
        self._vars_to_unfix = to_unfix
        self._var_was_fixed = None
        self._con_was_active = None
        self._comp_original_value = None
        self._var_was_unfixed = None
        self._remove_bounds_on_fix = remove_bounds_on_fix
        self._fixed_var_bounds = None

    def __enter__(self):
        to_fix = self._vars_to_fix
        to_deactivate = self._cons_to_deactivate
        to_set = self._comps_to_set
        to_unfix = self._vars_to_unfix
        self._var_was_fixed = [(var, var.fixed) for var in to_fix + to_unfix]
        self._con_was_active = [(con, con.active) for con in to_deactivate]
        self._comp_original_value = [(comp, comp.value) for comp in to_set]
        self._fixed_var_bounds = [(var.lb, var.ub) for var in to_fix]

        for var in self._vars_to_fix:
            if self._remove_bounds_on_fix:
                # TODO: Potentially override var.domain as well?
                var.setlb(None)
                var.setub(None)
            var.fix()

        for con in self._cons_to_deactivate:
            con.deactivate()

        for var in self._vars_to_unfix:
            # As of Pyomo 6.5, attempting to unfix an already unfixed var
            # does not raise an exception. Here we rely on this behavior.
            var.unfix()

        return self

    def __exit__(self, ex_type, ex_val, ex_bt):
        for var, was_fixed in self._var_was_fixed:
            if was_fixed:
                var.fix()
            else:
                var.unfix()
        if self._remove_bounds_on_fix:
            for var, (lb, ub) in zip(self._vars_to_fix, self._fixed_var_bounds):
                var.setlb(lb)
                var.setub(ub)

        for con, was_active in self._con_was_active:
            if was_active:
                con.activate()
        for comp, val in self._comp_original_value:
            comp.set_value(val)


class ParamSweeper(TemporarySubsystemManager):
    """This class enables setting values of variables/parameters
    according to a provided sequence. Iterating over this object
    sets values to the next in the sequence, at which point a
    calculation may be performed and output values compared.
    On exit, original values are restored.

    This is useful for testing a solve that is meant to perform some
    calculation, over a range of values for which the calculation
    is valid. For example:

    .. testcode::
       :skipif: not glpk_available

       model = pyo.ConcreteModel()
       model.v1 = pyo.Var()
       model.v2 = pyo.Var()
       model.c = pyo.Constraint(expr=model.v2 - model.v1 >= 0.1)
       model.o = pyo.Objective(expr=model.v1 + model.v2)
       solver = pyo.SolverFactory('glpk')
       input_vars = [model.v1]
       n_scen = 2
       input_values = pyo.ComponentMap([(model.v1, [1.1, 2.1])])
       output_values = pyo.ComponentMap([(model.v2, [1.2, 2.2])])
       with ParamSweeper(
               n_scen,
               input_values,
               output_values,
               to_fix=input_vars,
               ) as param_sweeper:
           for inputs, outputs in param_sweeper:
               solver.solve(model)
               # inputs and outputs contain the correct values for this
               # instance of the model
               for var, val in outputs.items():
                   # Test that model.v2 was calculated properly.
                   # First that it equals 1.2, then that it equals 2.2
                   assert var.value == val, f"{var.value} != {val}"

    """

    def __init__(
        self,
        n_scenario,
        input_values,
        output_values=None,
        to_fix=None,
        to_deactivate=None,
        to_reset=None,
    ):
        """
        Parameters
        ----------
        n_scenario: Integer
            The number of different values we expect for each input variable
        input_values: ComponentMap
            Maps each input variable to a list of values of length n_scenario
        output_values: ComponentMap
            Maps each output variable to a list of values of length n_scenario
        to_fix: List
            to_fix argument for base class
        to_deactivate: List
            to_deactivate argument for base class
        to_reset: List
            to_reset argument for base class. This list is extended with
            input variables.

        """
        # Should this object be aware of the user's block/model?
        # My answer for now is no.
        self.input_values = input_values
        output = ComponentMap() if output_values is None else output_values
        self.output_values = output
        self.n_scenario = n_scenario
        self.initial_state_values = None
        self._ip = -1  # Index pointer for iteration

        if to_reset is None:
            # Input values will be set repeatedly by iterating over this
            # object. Output values will presumably be altered by some
            # solve within this context. We would like to reset these to
            # their original values to make this functionality less
            # intrusive.
            to_reset = list(input_values) + list(output)
        else:
            to_reset.extend(var for var in input_values)
            to_reset.extend(var for var in output)

        super(ParamSweeper, self).__init__(
            to_fix=to_fix, to_deactivate=to_deactivate, to_reset=to_reset
        )

    def __iter__(self):
        return self

    def __next__(self):
        self._ip += 1

        i = self._ip
        n_scenario = self.n_scenario
        input_values = self.input_values
        output_values = self.output_values

        if i >= n_scenario:
            self._ip = -1
            raise StopIteration()

        else:
            inputs = ComponentMap()
            for var, values in input_values.items():
                val = values[i]
                var.set_value(val)
                inputs[var] = val

            outputs = ComponentMap(
                [(var, values[i]) for var, values in output_values.items()]
            )

            return inputs, outputs



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.benders.benders_cuts import BendersCutGenerator
import pyomo.environ as pyo


def create_root():
    m = pyo.ConcreteModel()
    m.y = pyo.Var(bounds=(1, None))
    m.eta = pyo.Var(bounds=(-10, None))
    m.obj = pyo.Objective(expr=m.y**2 + m.eta)
    return m


def create_subproblem(root):
    m = pyo.ConcreteModel()
    m.x1 = pyo.Var()
    m.x2 = pyo.Var()
    m.y = pyo.Var()
    m.obj = pyo.Objective(expr=-m.x2)
    m.c1 = pyo.Constraint(expr=(m.x1 - 1) ** 2 + m.x2**2 <= pyo.log(m.y))
    m.c2 = pyo.Constraint(expr=(m.x1 + 1) ** 2 + m.x2**2 <= pyo.log(m.y))

    complicating_vars_map = pyo.ComponentMap()
    complicating_vars_map[root.y] = m.y

    return m, complicating_vars_map


def main():
    m = create_root()
    root_vars = [m.y]
    m.benders = BendersCutGenerator()
    m.benders.set_input(root_vars=root_vars, tol=1e-8)
    m.benders.add_subproblem(
        subproblem_fn=create_subproblem,
        subproblem_fn_kwargs={'root': m},
        root_eta=m.eta,
        subproblem_solver='ipopt',
    )
    opt = pyo.SolverFactory('gurobi_direct')

    for i in range(30):
        res = opt.solve(m, tee=False)
        cuts_added = m.benders.generate_cut()
        print(len(cuts_added), m.y.value, m.eta.value)
        if len(cuts_added) == 0:
            break


if __name__ == '__main__':
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import mpi4py
from pyomo.contrib.pynumero.sparse import BlockVector
from .base_block import (
    BaseBlockVector,
    vec_unary_ufuncs,
    vec_binary_ufuncs,
    vec_associative_reductions,
)
from .block_vector import NotFullyDefinedBlockVectorError
from .block_vector import assert_block_structure as block_vector_assert_block_structure
import numpy as np
import operator


def assert_block_structure(vec):
    if vec.has_none:
        msg = 'Call MPIBlockVector.broadcast_block_sizes() first.'
        raise NotFullyDefinedBlockVectorError(msg)


class MPIBlockVector(BaseBlockVector, np.ndarray):
    """
    Parallel structured vector interface. This interface can be used to
    perform parallel operations on vectors composed by vectors. The main
    idea is to allocate vectors in different processors and make the corresponding
    parallel calls when necessary.

    Attributes
    ----------
    _rank_owner: numpy.ndarray
        1D-array with processor ownership of each block. A block can be own by a
        single processor or by all processors. Blocks own by all processors have
        ownership -1. Blocks own by a single processor have ownership rank. where
        rank=MPI.COMM_WORLD.Get_rank()
    _mpiw: MPI communicator
        A communicator from the MPI space. Typically MPI.COMM_WORLD
    _block_vector: BlockVector
        Internal BlockVector. Blocks that belong to this processor are stored
        in _block_vector. Blocks that do not belong to this processor are empty
        and store as numpy.zeros(0)
    _owned_mask: numpy.ndarray bool
        1D-array that indicates if a block belongs to this processor. While
        _rank_owner tells which processor(s) owns each block, _owned_mask tells
        if a block is owned by this processor. Blocks that are owned by everyone
        (i.e. ownership = -1) are True in _owned_mask
    _owned_blocks: numpy.ndarray
        1D-array with block indices owned by this processor. This includes blocks
        with ownership -1.
    _unique_owned_blocks: numpy.ndarray
        1D-array with block indices owned only by this processor. This does not
        include blocks with ownership -1.
    _brow_lengths: numpy.ndarray
        1D-Array of size nblocks that specifies the length of each entry
        in the MPIBlockVector. This is the same across all processors.
    _undefined_brows: set
        A set of block indices for which the blocks are still None (i.e., the dimensions
        have not yet ben set). Operations with BlockVectors require all entries to be
        different than None.

    Notes
    ------
    This is the parallel implementation of pyomo.contrib.pynumero.sparse.BlockVector

    Parameters
    -------------------
    nblocks: int
        number of blocks contained in the block vector
    rank_owner: array_like
        Array_like of size nblocks. Each entry defines ownership of each block.
        There are two types of ownership. Block that are owned by all processor,
        and blocks owned by a single processor. If a block is owned by all
        processors then its ownership is -1. Otherwise, if a block is owned by
        a single processor, then its ownership is equal to the rank of the
        processor.
    mpi_com: MPI communicator
        An MPI communicator. Tyically MPI.COMM_WORLD
    """

    def __new__(cls, nblocks, rank_owner, mpi_comm, assert_correct_owners=False):
        assert isinstance(nblocks, int)
        assert len(rank_owner) == nblocks

        blocks = [None for i in range(nblocks)]
        arr = np.asarray(blocks, dtype='object')
        obj = arr.view(cls)

        obj._rank_owner = np.array([i for i in rank_owner])
        obj._mpiw = mpi_comm
        obj._block_vector = BlockVector(nblocks)

        rank = obj._mpiw.Get_rank()
        comm_size = obj._mpiw.Get_size()
        assert np.all(obj._rank_owner < comm_size)

        # Determine which blocks are owned by this processor
        obj._owned_mask = np.bitwise_or(obj._rank_owner == rank, obj._rank_owner < 0)
        unique_owned_mask = obj._rank_owner == rank
        obj._unique_owned_blocks = unique_owned_mask.nonzero()[0]
        obj._owned_blocks = obj._owned_mask.nonzero()[0]

        # containers that facilitate looping
        obj._brow_lengths = np.empty(nblocks, dtype=np.float64)
        obj._brow_lengths.fill(np.nan)
        obj._undefined_brows = set(obj._owned_blocks)

        # make some pointers unmutable. These arrays don't change after
        # MPIBlockVector has been created
        obj._rank_owner.flags.writeable = False
        obj._owned_blocks.flags.writeable = False
        obj._owned_mask.flags.writeable = False
        obj._unique_owned_blocks.flags.writeable = False

        obj._broadcasted = False

        return obj

    def __init__(self, nblocks, rank_owner, mpi_comm, assert_correct_owners=False):
        # Note: this requires communication but is disabled when assertions
        # are turned off
        if assert_correct_owners:
            assert (
                self._assert_correct_owners()
            ), 'rank_owner must be the same in all processors'

    def __array_prepare__(self, out_arr, context=None):
        return super(MPIBlockVector, self).__array_prepare__(self, out_arr, context)

    def __array_wrap__(self, out_arr, context=None):
        return super(MPIBlockVector, self).__array_wrap__(self, out_arr, context)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Runs ufuncs speciallizations to MPIBlockVector"""
        outputs = kwargs.pop('out', None)
        if outputs is not None:
            raise NotImplementedError(
                str(ufunc)
                + ' cannot be used with MPIBlockVector if the out keyword argument is given.'
            )

        if ufunc in vec_unary_ufuncs:
            results = self._unary_operation(ufunc, method, *inputs, **kwargs)
            return results
        elif ufunc in vec_binary_ufuncs:
            results = self._binary_operation(ufunc, method, *inputs, **kwargs)
            return results
        else:
            raise NotImplementedError(str(ufunc) + "not supported for MPIBlockVector")

    def _unary_operation(self, ufunc, method, *args, **kwargs):
        """Run recursion to perform unary_funcs on MPIBlockVector"""
        # ToDo: deal with out
        x = args[0]

        if isinstance(x, MPIBlockVector):
            rank = self._mpiw.Get_rank()
            v = x.copy_structure()
            for i in self._owned_blocks:
                _args = [x.get_block(i)] + [args[j] for j in range(1, len(args))]
                v.set_block(i, self._unary_operation(ufunc, method, *_args, **kwargs))
            return v
        elif isinstance(x, BlockVector):
            v = BlockVector(x.nblocks)
            for i in range(x.nblocks):
                _args = [x.get_block(i)] + [args[j] for j in range(1, len(args))]
                v.set_block(i, self._unary_operation(ufunc, method, *_args, **kwargs))
            return v
        elif type(x) == np.ndarray:
            return super(MPIBlockVector, self).__array_ufunc__(
                ufunc, method, *args, **kwargs
            )
        else:
            raise NotImplementedError()

    def _binary_operation(self, ufunc, method, *args, **kwargs):
        """Run recursion to perform binary_funcs on MPIBlockVector"""
        # ToDo: deal with out
        x1 = args[0]
        x2 = args[1]
        if isinstance(x1, MPIBlockVector) and isinstance(x2, MPIBlockVector):
            msg = 'BlockVectors must be distributed in same processors'
            assert (
                np.array_equal(x1._rank_owner, x2._rank_owner)
                or self._mpiw.Get_size() == 1
            ), msg
            assert x1._mpiw == x2._mpiw, 'Need to have same communicator'

            res = x1.copy_structure()
            for i in x1._owned_blocks:
                _args = (
                    [x1.get_block(i)]
                    + [x2.get_block(i)]
                    + [args[j] for j in range(2, len(args))]
                )
                res.set_block(
                    i, self._binary_operation(ufunc, method, *_args, **kwargs)
                )
            return res
        elif isinstance(x1, BlockVector) and isinstance(x2, MPIBlockVector):
            raise RuntimeError('Operation not supported by MPIBlockVector')
        elif isinstance(x1, MPIBlockVector) and isinstance(x2, BlockVector):
            raise RuntimeError('Operation not supported by MPIBlockVector')
        elif isinstance(x1, MPIBlockVector) and np.isscalar(x2):
            res = x1.copy_structure()
            for i in x1._owned_blocks:
                _args = (
                    [x1.get_block(i)] + [x2] + [args[j] for j in range(2, len(args))]
                )
                res.set_block(
                    i, self._binary_operation(ufunc, method, *_args, **kwargs)
                )
            return res
        elif isinstance(x2, MPIBlockVector) and np.isscalar(x1):
            res = x2.copy_structure()
            for i in x2._owned_blocks:
                _args = (
                    [x1] + [x2.get_block(i)] + [args[j] for j in range(2, len(args))]
                )
                res.set_block(
                    i, self._binary_operation(ufunc, method, *_args, **kwargs)
                )
            return res
        elif isinstance(x1, MPIBlockVector) and type(x2) == np.ndarray:
            raise RuntimeError('Operation not supported by MPIBlockVector')
        elif isinstance(x2, MPIBlockVector) and type(x1) == np.ndarray:
            raise RuntimeError('Operation not supported by MPIBlockVector')
        elif isinstance(x1, np.ndarray) and isinstance(x2, np.ndarray):
            # this will take care of blockvector and ndarrays
            return self._block_vector.__array_ufunc__(ufunc, method, *args, **kwargs)
        elif (type(x1) == BlockVector or np.isscalar(x1)) and (
            type(x2) == BlockVector or np.isscalar(x2)
        ):
            return self._block_vector.__array_ufunc__(ufunc, method, *args, **kwargs)
        elif (type(x1) == np.ndarray or np.isscalar(x1)) and (
            type(x2) == np.ndarray or np.isscalar(x2)
        ):
            return super(MPIBlockVector, self).__array_ufunc__(
                ufunc, method, *args, **kwargs
            )
        else:
            raise NotImplementedError()

    @property
    def nblocks(self):
        """
        Returns the number of blocks.
        """
        return self._block_vector.nblocks

    @property
    def bshape(self):
        """
        Returns the number of blocks in this MPIBlockVector in a tuple.
        """
        return (self.nblocks,)

    @property
    def shape(self):
        """
        Returns total number of elements in the MPIBlockVector
        """
        return (self.size,)

    @property
    def size(self):
        """
        Returns total number of elements in this MPIBlockVector
        """
        assert_block_structure(self)
        comm = self._mpiw
        rank = comm.Get_rank()
        if rank == 0:
            indices = self._owned_blocks
        else:
            indices = self._unique_owned_blocks
        local_size = np.sum(self._brow_lengths[indices])
        size = comm.allreduce(local_size)
        assert int(size) == size
        return int(size)

    @property
    def ndim(self):
        """
        Returns dimension of this MPIBlockVector
        """
        return 1

    @property
    def has_none(self):
        """
        Returns True if block vector has none entry
        """
        return len(self._undefined_brows) != 0

    @property
    def owned_blocks(self):
        """
        Returns list with indices of blocks owned by this processor.
        """
        return self._owned_blocks

    @property
    def shared_blocks(self):
        """
        Returns list with indices of blocks shared by all processors
        """
        return np.array([i for i in range(self.nblocks) if self._rank_owner[i] < 0])

    @property
    def rank_ownership(self):
        """
        Returns 1D-Array with processor ranks that own each block. The ownership
        of blocks that are owned by all processors is -1.
        """
        return self._rank_owner

    @property
    def ownership_mask(self):
        """
        Returns boolean 1D-Array that indicates which blocks are owned by
        this processor
        """
        return self._owned_mask

    @property
    def mpi_comm(self):
        """Returns MPI communicator"""
        return self._mpiw

    def is_broadcasted(self):
        return self._broadcasted

    def block_sizes(self, copy=True):
        """
        Returns 1D-Array with sizes of individual blocks in this MPIBlockVector
        """
        if not self._broadcasted:
            self.broadcast_block_sizes()
        if copy:
            return self._brow_lengths.copy()
        return self._brow_lengths

    def get_block_size(self, ndx):
        res = self._brow_lengths[ndx]
        if np.isnan(res):
            raise NotFullyDefinedBlockVectorError(
                'The dimensions of the requested block are not defined.'
            )
        res = int(res)
        return res

    def _set_block_size(self, ndx, size):
        if ndx in self._undefined_brows:
            self._undefined_brows.remove(ndx)
            self._brow_lengths[ndx] = size
        else:
            if self._brow_lengths[ndx] != size:
                raise ValueError(
                    'Incompatible dimensions for block {ndx}; '
                    'got {got}; expected {exp}'.format(
                        ndx=ndx, got=size, exp=self._brow_lengths[ndx]
                    )
                )

    # Note: this operation requires communication
    def broadcast_block_sizes(self):
        """
        Send sizes of all blocks to all processors. After this method is called
        this MPIBlockVector knows it's dimensions across all blocks. This method
        must be called before running any operations with the MPIBlockVector.
        """
        assert_block_structure(self)
        rank = self._mpiw.Get_rank()
        num_processors = self._mpiw.Get_size()

        local_length_data = np.empty(self.nblocks, dtype=np.int64)
        local_length_data.fill(-1)
        for ndx in self.owned_blocks:
            if ndx in self._undefined_brows:
                raise NotFullyDefinedBlockVectorError(
                    'Block {ndx} is owned by rank {rank}, '
                    'but the dimensions for block {ndx} '
                    'have not yet been specified in rank {rank}. '
                    'Please specify all owned blocks.'.format(ndx=ndx, rank=rank)
                )
            local_length_data[ndx] = self.get_block_size(ndx)
        receive_data = np.empty(num_processors * self.nblocks, dtype=np.int64)
        self._mpiw.Allgather(local_length_data, receive_data)

        proc_dims = np.split(receive_data, num_processors)

        for i in range(self.nblocks):
            block_length = set()
            for k in range(num_processors):
                processor_sizes = proc_dims[k]
                block_length.add(processor_sizes[i])
            if len(block_length) > 2:
                msg = 'Block {} has more than one dimension across processors'.format(i)
                raise RuntimeError(msg)
            elif len(block_length) == 2:
                if -1 not in block_length:
                    msg = (
                        'Block {} has more than one dimension across processors'.format(
                            i
                        )
                    )
                    raise RuntimeError(msg)
                block_length.remove(-1)
            elif -1 in block_length:
                msg = (
                    'The dimension of block {} was not specified in any process'.format(
                        i
                    )
                )

            # here block_length must only have one element
            self._brow_lengths[i] = block_length.pop()

        self._brow_lengths = np.asarray(self._brow_lengths, dtype=np.int64)
        self._broadcasted = True

    def finalize_block_sizes(self, broadcast=True, block_sizes=None):
        """
        Only set broadcast=False if you know what you are doing!

        Parameters
        ----------
        broadcast: bool
        block_sizes: None or np.ndarray
        """
        if broadcast:
            self.broadcast_block_sizes()
        else:
            self._undefined_brows = set()
            self._brow_lengths = block_sizes
            self._broadcasted = True

    # Note: this requires communication but is only run in __new__
    def _assert_correct_owners(self, root=0):
        rank = self._mpiw.Get_rank()
        num_processors = self._mpiw.Get_size()

        if num_processors == 1:
            return True

        local_owners = self._rank_owner.copy()
        receive_data = None
        if rank == root:
            receive_data = np.empty(self.nblocks * num_processors, dtype=np.int64)

        self._mpiw.Gather(local_owners, receive_data, root=root)

        if rank == root:
            owners_in_processor = np.split(receive_data, num_processors)
            root_rank_owners = owners_in_processor[root]
            for i in range(self.nblocks):
                for k in range(num_processors):
                    if k != root:
                        if owners_in_processor[k][i] != root_rank_owners[i]:
                            return False
        return True

    def all(self, axis=None, out=None, keepdims=False):
        """
        Returns True if all elements evaluate to True.
        """
        assert out is None, 'Out keyword not supported'
        assert_block_structure(self)
        local = 1
        for i in self._owned_blocks:
            local *= self._block_vector.get_block(i).all()

        return bool(self._mpiw.allreduce(local, op=mpi4py.MPI.PROD))

    def any(self, axis=None, out=None, keepdims=False):
        """
        Returns True if all elements evaluate to True.
        """
        assert out is None, 'Out keyword not supported'
        assert_block_structure(self)
        local = 0
        for i in self._owned_blocks:
            local += self._block_vector.get_block(i).any()

        return bool(self._mpiw.allreduce(local, op=mpi4py.MPI.SUM))

    def min(self, axis=None, out=None, keepdims=False):
        """
        Returns the smallest value stored in the vector
        """
        assert out is None, 'Out keyword not supported'
        assert_block_structure(self)
        local_min = np.inf
        for i in self._owned_blocks:
            block = self._block_vector.get_block(i)
            if block.size > 0:
                lmin = block.min()
                if lmin <= local_min:
                    local_min = lmin
        res = self._mpiw.allreduce(local_min, op=mpi4py.MPI.MIN)
        if res == np.inf:
            if self.size == 0:
                raise ValueError('cannot get the min of a size 0 array')
        return res

    def max(self, axis=None, out=None, keepdims=False):
        """
        Returns the largest value stored in this MPIBlockVector
        """
        assert out is None, 'Out keyword not supported'
        assert_block_structure(self)
        local_max = -np.inf
        for i in self._owned_blocks:
            block = self._block_vector.get_block(i)
            if block.size > 0:
                lmax = block.max()
                if lmax >= local_max:
                    local_max = lmax
        res = self._mpiw.allreduce(local_max, op=mpi4py.MPI.MAX)
        if res == -np.inf:
            if self.size == 0:
                raise ValueError('cannot get the max of a size 0 array')
        return res

    def sum(self, axis=None, dtype=None, out=None, keepdims=False):
        """
        Returns the sum of all entries in this MPIBlockVector
        """
        assert out is None, 'Out keyword not supported'
        assert_block_structure(self)
        rank = self._mpiw.Get_rank()
        indices = self._unique_owned_blocks if rank != 0 else self._owned_blocks

        local_sum = 0.0
        for i in indices:
            local_sum += self._block_vector.get_block(i).sum(axis=axis, dtype=dtype)

        return self._mpiw.allreduce(local_sum, op=mpi4py.MPI.SUM)

    def prod(self, axis=None, dtype=None, out=None, keepdims=False):
        """
        Returns the product of all entries in this MPIBlockVector
        """
        assert out is None, 'Out keyword not supported'
        assert_block_structure(self)
        rank = self._mpiw.Get_rank()
        indices = self._unique_owned_blocks if rank != 0 else self._owned_blocks

        local_prod = 1.0
        for i in indices:
            local_prod *= self._block_vector.get_block(i).prod(axis=axis, dtype=dtype)
        return self._mpiw.allreduce(local_prod, op=mpi4py.MPI.PROD)

    def mean(self, axis=None, dtype=None, out=None, keepdims=False):
        """
        Returns the average of all entries in this MPIBlockVector
        """
        return self.sum(out=out) / self.size

    def conj(self):
        """
        Complex-conjugate all elements.
        """
        assert_block_structure(self)
        result = self.copy_structure()
        for i in self._owned_blocks:
            result.set_block(i, self.get_block(i).conj())
        return result

    def conjugate(self):
        """
        Complex-conjugate all elements.
        """
        return self.conj()

    def nonzero(self):
        """
        Returns the indices of the elements that are non-zero.
        """
        result = MPIBlockVector(
            nblocks=self.nblocks,
            rank_owner=self.rank_ownership,
            mpi_comm=self.mpi_comm,
            assert_correct_owners=False,
        )
        assert_block_structure(self)
        for i in self._owned_blocks:
            result.set_block(i, self._block_vector.get_block(i).nonzero()[0])
        return (result,)

    def round(self, decimals=0, out=None):
        """
        Return MPIBlockVector with each element rounded to the given number of decimals
        """
        assert out is None, 'Out keyword not supported'
        assert_block_structure(self)
        result = self.copy_structure()
        for i in self._owned_blocks:
            result.set_block(
                i, self._block_vector.get_block(i).round(decimals=decimals)
            )
        return result

    def clip(self, min=None, max=None, out=None):
        """
        Return MPIBlockVector whose values are limited to [min, max].
        One of max or min must be given.

        Parameters
        ----------
        min: scalar_like, optional
            Minimum value. If None, clipping is not performed on lower interval edge.
        max: scalar_like, optional
            Maximum value. If None, clipping is not performed on upper interval edge.

        Returns
        -------
        MPIBlockVector

        """
        assert out is None, 'Out keyword not supported'
        assert_block_structure(self)
        result = self.copy_structure()
        for i in self._owned_blocks:
            result.set_block(i, self._block_vector.get_block(i).clip(min=min, max=max))
        return result

    def compress(self, condition, axis=None, out=None):
        """
        Return selected slices of each subblock.

        Parameters
        ----------
        condition: MPIBlockVector that selects which entries to return.
            Determines to select (evaluate True in condition)

        Returns
        -------
        MPIBlockVector

        """
        assert out is None, 'Out keyword not supported'
        assert_block_structure(self)
        result = MPIBlockVector(
            nblocks=self.nblocks,
            rank_owner=self.rank_ownership,
            mpi_comm=self.mpi_comm,
            assert_correct_owners=False,
        )
        if isinstance(condition, MPIBlockVector):
            # Note: do not need to check same size? this is checked implicitly
            msg = 'BlockVectors must be distributed in same processors'
            assert np.array_equal(self._rank_owner, condition._rank_owner), msg
            assert self._mpiw == condition._mpiw, 'Need to have same communicator'
            for i in self._owned_blocks:
                result.set_block(i, self.get_block(i).compress(condition.get_block(i)))
            return result
        if isinstance(condition, BlockVector):
            raise RuntimeError('Operation not supported by MPIBlockVector')
        elif isinstance(condition, np.ndarray):
            raise RuntimeError('Operation not supported by MPIBlockVector')
        else:
            raise NotImplementedError('Operation not supported by MPIBlockVector')

    def copyfrom(self, other):
        """
        Copy entries of other into this MPIBlockVector

        Parameters
        ----------
        other: MPIBlockVector or BlockVector

        Returns
        -------
        None
        """
        if isinstance(other, MPIBlockVector):
            assert_block_structure(other)
            msg = 'Number of blocks mismatch {} != {}'.format(
                self.nblocks, other.nblocks
            )
            assert self.nblocks == other.nblocks, msg
            msg = 'BlockVectors must be distributed in same processors'
            assert np.array_equal(self._rank_owner, other.rank_ownership), msg
            assert self._mpiw == other._mpiw, 'Need to have same communicator'

            for i in self._owned_blocks:
                self.set_block(i, other.get_block(i).copy())

        elif isinstance(other, BlockVector):
            block_vector_assert_block_structure(other)
            msg = 'Number of blocks mismatch {} != {}'.format(
                self.nblocks, other.nblocks
            )
            assert self.nblocks == other.nblocks, msg
            for i in self._owned_blocks:
                self.set_block(i, other.get_block(i).copy())
        elif isinstance(other, np.ndarray):
            assert_block_structure(self)
            if not self.is_broadcasted():
                self.broadcast_block_sizes()
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            offset = 0
            for idx in range(self.nblocks):
                if self._owned_mask[idx]:
                    subarray = other[offset : offset + self.get_block_size(idx)]
                    if isinstance(self.get_block(idx), BlockVector):
                        self.get_block(idx).copyfrom(subarray)
                    else:
                        np.copyto(self.get_block(idx), subarray)
                offset += self.get_block_size(idx)
        else:
            raise NotImplementedError('Operation not supported by MPIBlockVector')

    def copyto(self, other):
        """
        Copy entries of this MPIBlockVector into other

        Parameters
        ----------
        other: MPIBlockVector or BlockVector

        Returns
        -------
        None

        """
        if isinstance(other, MPIBlockVector):
            other.copyfrom(self)
        else:
            raise NotImplementedError('Operation not supported by MPIBlockVector')

    def set_blocks(self, blocks):
        """
        Assigns vectors in blocks

        Parameters
        ----------
        blocks: list
            list of vectors

        Returns
        -------
        None
        """
        raise NotImplementedError('Operation not supported by MPIBlockVector')

    def clone(self, value=None, copy=True):
        """
        Returns a copy of this MPIBlockVector

        Parameters
        ----------
        value: scalar, optional
            all entries of the cloned vector are set to this value
        copy: bool, optional
            if set to true makes a deepcopy of each block in this vector. default False

        Returns
        -------
        MPIBlockVector
        """
        result = MPIBlockVector(
            self.nblocks,
            self.rank_ownership,
            self.mpi_comm,
            assert_correct_owners=False,
        )
        result._block_vector = self._block_vector.clone(copy=copy)
        result._brow_lengths = self._brow_lengths.copy()
        result._undefined_brows = set(self._undefined_brows)
        if value is not None:
            result.fill(value)
        return result

    def copy(self, order='C'):
        """
        Returns a copy of the MPIBlockVector
        """
        result = MPIBlockVector(
            self.nblocks,
            self.rank_ownership,
            self.mpi_comm,
            assert_correct_owners=False,
        )
        result._block_vector = self._block_vector.copy(order=order)
        result._brow_lengths = self._brow_lengths.copy()
        result._undefined_brows = set(self._undefined_brows)
        return result

    def copy_structure(self):
        """
        Returns a copy of the MPIBlockVector structure filled with zeros
        """
        result = MPIBlockVector(
            self.nblocks,
            self.rank_ownership,
            self.mpi_comm,
            assert_correct_owners=False,
        )
        if self.is_broadcasted():
            result.finalize_block_sizes(
                broadcast=False, block_sizes=self.block_sizes(copy=False)
            )
        for bid in self.owned_blocks:
            block = self.get_block(bid)
            if block is not None:
                if isinstance(block, BlockVector):
                    result.set_block(bid, block.copy_structure())
                elif type(block) == np.ndarray:
                    result.set_block(bid, np.zeros(block.size))
                else:
                    raise NotImplementedError('Should never get here')
        return result

    def fill(self, value):
        """
        Fills the MPIBLockVector with a scalar value.

        Parameters
        ----------
        value : scalar
            All elements in the vector will be assigned this value

        Returns
        -------
        None

        """
        assert_block_structure(self)
        for idx in self.owned_blocks:
            self.get_block(idx).fill(value)

    def dot(self, other, out=None):
        """
        Returns dot product

        Parameters
        ----------
        other : MPIBlockVector

        Returns
        -------
        float

        """
        assert_block_structure(self)
        assert out is None
        if isinstance(other, MPIBlockVector):
            assert_block_structure(other)
            msg = 'BlockVectors must be distributed in same processors'
            assert np.array_equal(self.rank_ownership, other.rank_ownership), msg
            assert self.mpi_comm == other.mpi_comm, 'Need to have same communicator'

            rank = self._mpiw.Get_rank()
            indices = self._unique_owned_blocks if rank != 0 else self._owned_blocks
            local_dot_prod = 0.0
            for i in indices:
                local_dot_prod += self._block_vector.get_block(i).dot(
                    other.get_block(i)
                )

            return self._mpiw.allreduce(local_dot_prod, op=mpi4py.MPI.SUM)
        elif isinstance(other, BlockVector):
            assert (
                self.nblocks == other.nblocks
            ), 'Number of blocks mismatch: {} != {}'.format(self.nblocks, other.nblocks)
            return self.dot(
                other.toMPIBlockVector(
                    self.rank_ownership, self.mpi_comm, assert_correct_owners=False
                )
            )
        elif isinstance(other, np.ndarray):
            other_bv = self.copy_structure()
            other_bv.copyfrom(other)
            return self.dot(other_bv)
        else:
            raise NotImplementedError('Operation not supported by MPIBlockVector')

    @staticmethod
    def _serialize_structure(block_vector):
        """
        Parameters
        ----------
        block_vector: BlockVector

        Returns
        -------
        list
        """
        serialized_structure = list()
        for ndx in range(block_vector.nblocks):
            blk = block_vector.get_block(ndx)
            if isinstance(blk, BlockVector):
                serialized_structure.append(-1)
                serialized_structure.append(blk.nblocks)
                serialized_structure.extend(MPIBlockVector._serialize_structure(blk))
            elif isinstance(blk, MPIBlockVector):
                raise NotImplementedError(
                    'Operation not supported for MPIBlockVectors containing other MPIBlockVectors'
                )
            elif isinstance(blk, np.ndarray):
                serialized_structure.append(-2)
                serialized_structure.append(blk.size)
            else:
                raise NotImplementedError('Unrecognized input.')
        return serialized_structure

    @staticmethod
    def _create_from_serialized_structure(serialized_structure, structure_ndx, result):
        """
        Parameters
        ----------
        serialized_structure: np.ndarray
        structure_ndx: int
        result: BlockVector

        Returns
        -------
        structure_ndx: int
        """
        for ndx in range(result.nblocks):
            if serialized_structure[structure_ndx] == -1:
                structure_ndx += 1
                block = BlockVector(serialized_structure[structure_ndx])
                structure_ndx += 1
                structure_ndx = MPIBlockVector._create_from_serialized_structure(
                    serialized_structure, structure_ndx, block
                )
                result.set_block(ndx, block)
            elif serialized_structure[structure_ndx] == -2:
                structure_ndx += 1
                result.set_block(ndx, np.zeros(serialized_structure[structure_ndx]))
                structure_ndx += 1
            else:
                raise ValueError('Unrecognized structure')
        return structure_ndx

    def make_local_structure_copy(self):
        """
        Creates a BlockVector with the same structure as the MPIBlockVector

        Returns
        -------
        BlockVector
        """
        """
        We do this by serializing the structure, then gathering it.
        To serialize the structure, we use an array. The first number indicates if the first block is a numpy array 
        or a BlockVector. We use -1 to indicate a BlockVector and -2 to indicate a numpy array. If the block is a
        BlockVector, then the next number is a positive integer specifying the number of blocks in the block vector.
        If the block is a numpy array, then the next number is a positive integer specifying the size of the array. 
        After the number of blocks in a BlockVector is specified, we follow the same procedure to specify the 
        structure of that BlockVector. 
        """
        assert_block_structure(self)
        serialized_structure_by_block = dict()
        length_per_block = np.zeros(self.nblocks, dtype=np.int64)
        rank = self._mpiw.Get_rank()
        if rank == 0:
            block_indices = self._owned_blocks
        else:
            block_indices = self._unique_owned_blocks
        for ndx in block_indices:
            blk = self.get_block(ndx)
            blk_structure = list()
            if isinstance(blk, BlockVector):
                blk_structure.append(-1)
                blk_structure.append(blk.nblocks)
                blk_structure.extend(self._serialize_structure(blk))
            elif isinstance(blk, MPIBlockVector):
                raise NotImplementedError(
                    'Operation not supported for MPIBlockVectors containing other MPIBlockVectors'
                )
            elif isinstance(blk, np.ndarray):
                blk_structure.append(-2)
                blk_structure.append(blk.size)
            else:
                raise NotImplementedError('Unrecognized input.')
            length_per_block[ndx] = len(blk_structure)
            serialized_structure_by_block[ndx] = np.asarray(
                blk_structure, dtype=np.int64
            )

        global_length_per_block = np.zeros(self.nblocks, dtype=np.int64)
        self._mpiw.Allreduce(length_per_block, global_length_per_block)
        local_serialized_structure = np.zeros(
            global_length_per_block.sum(), dtype=np.int64
        )

        offset = 0
        block_indices_set = set(block_indices)
        for ndx in range(self.nblocks):
            if ndx in block_indices_set:
                local_serialized_structure[
                    offset : offset + global_length_per_block[ndx]
                ] = serialized_structure_by_block[ndx]
            offset += global_length_per_block[ndx]
        global_serialized_structure = np.zeros(
            global_length_per_block.sum(), dtype=np.int64
        )
        self._mpiw.Allreduce(local_serialized_structure, global_serialized_structure)

        result = BlockVector(self.nblocks)
        structure_ndx = 0
        self._create_from_serialized_structure(
            global_serialized_structure, structure_ndx, result
        )

        return result

    def make_local_copy(self):
        """
        Copies the MPIBlockVector into a BlockVector

        Returns
        -------
        BlockVector
        """
        assert_block_structure(self)
        if not self.is_broadcasted():
            self.broadcast_block_sizes()
        result = self.make_local_structure_copy()

        local_data = np.zeros(self.size)
        global_data = np.zeros(self.size)

        offset = 0
        rank = self._mpiw.Get_rank()
        if rank == 0:
            block_indices = set(self._owned_blocks)
        else:
            block_indices = set(self._unique_owned_blocks)
        for ndx in range(self.nblocks):
            if ndx in block_indices:
                blk = self.get_block(ndx)
                if isinstance(blk, BlockVector):
                    local_data[offset : offset + self.get_block_size(ndx)] = (
                        blk.flatten()
                    )
                elif isinstance(blk, np.ndarray):
                    local_data[offset : offset + self.get_block_size(ndx)] = blk
                else:
                    raise ValueError('Unrecognized block type')
            offset += self.get_block_size(ndx)

        self._mpiw.Allreduce(local_data, global_data)
        result.copyfrom(global_data)

        return result

    def _binary_operation_helper(self, other, operation):
        assert_block_structure(self)
        result = self.copy_structure()
        if isinstance(other, MPIBlockVector) or isinstance(other, BlockVector):
            assert (
                self.nblocks == other.nblocks
            ), 'Number of blocks mismatch: {} != {}'.format(self.nblocks, other.nblocks)
            if isinstance(other, MPIBlockVector):
                assert (
                    np.array_equal(self._rank_owner, other._rank_owner)
                    or self._mpiw.Get_size() == 1
                ), 'MPIBlockVectors must be distributed in same processors'
                assert self._mpiw == other._mpiw, 'Need to have same communicator'
            for i in self._owned_blocks:
                result.set_block(i, operation(self.get_block(i), other.get_block(i)))
            return result
        elif isinstance(other, np.ndarray):
            _tmp = self.copy_structure()
            _tmp.copyfrom(other)
            return self._binary_operation_helper(_tmp, operation)
        elif np.isscalar(other):
            for i in self._owned_blocks:
                result.set_block(i, operation(self.get_block(i), other))
            return result
        else:
            raise NotImplementedError('Operation not supported by MPIBlockVector')

    def _reverse_binary_operation_helper(self, other, operation):
        assert_block_structure(self)
        result = self.copy_structure()
        if isinstance(other, BlockVector):
            raise RuntimeError('Operation not supported by MPIBlockVector')
        elif isinstance(other, np.ndarray):
            raise RuntimeError('Operation not supported by MPIBlockVector')
        elif np.isscalar(other):
            for i in self._owned_blocks:
                result.set_block(i, operation(other, self.get_block(i)))
            return result
        else:
            raise NotImplementedError('Operation not supported by MPIBlockVector')

    def _inplace_binary_operation_helper(self, other, operation):
        assert_block_structure(self)
        if isinstance(other, MPIBlockVector) or isinstance(other, BlockVector):
            assert (
                self.nblocks == other.nblocks
            ), 'Number of blocks mismatch: {} != {}'.format(self.nblocks, other.nblocks)
            if isinstance(other, MPIBlockVector):
                assert (
                    np.array_equal(self._rank_owner, other._rank_owner)
                    or self._mpiw.Get_size() == 1
                ), 'MPIBlockVectors must be distributed in same processors'
                assert self._mpiw == other._mpiw, 'Need to have same communicator'
                assert_block_structure(other)
            else:
                block_vector_assert_block_structure(other)

            for i in self._owned_blocks:
                blk = self.get_block(i)
                operation(blk, other.get_block(i))
                self.set_block(i, blk)
            return self
        elif isinstance(other, np.ndarray):
            _tmp = self.copy_structure()
            _tmp.copyfrom(other)
            return self._inplace_binary_operation_helper(_tmp, operation)
        elif np.isscalar(other):
            for i in self._owned_blocks:
                blk = self.get_block(i)
                operation(blk, other)
                self.set_block(i, blk)
            return self
        else:
            raise NotImplementedError('Operation not supported by MPIBlockVector')

    def __add__(self, other):
        return self._binary_operation_helper(other, operator.add)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self._binary_operation_helper(other, operator.sub)

    def __rsub__(self, other):
        return self._reverse_binary_operation_helper(other, operator.sub)

    def __mul__(self, other):
        return self._binary_operation_helper(other, operator.mul)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return self._binary_operation_helper(other, operator.truediv)

    def __rtruediv__(self, other):
        return self._reverse_binary_operation_helper(other, operator.truediv)

    def __floordiv__(self, other):
        return self._binary_operation_helper(other, operator.floordiv)

    def __rfloordiv__(self, other):
        return self._reverse_binary_operation_helper(other, operator.floordiv)

    def __neg__(self):
        assert_block_structure(self)
        result = self.copy_structure()
        for ndx in self._owned_blocks:
            result.set_block(ndx, -self.get_block(ndx))
        return result

    def __iadd__(self, other):
        return self._inplace_binary_operation_helper(other, operator.iadd)

    def __isub__(self, other):
        return self._inplace_binary_operation_helper(other, operator.isub)

    def __imul__(self, other):
        return self._inplace_binary_operation_helper(other, operator.imul)

    def __itruediv__(self, other):
        return self._inplace_binary_operation_helper(other, operator.itruediv)

    def __div__(self, other):
        return self.__truediv__(other)

    def __rdiv__(self, other):
        return self.__rtruediv__(other)

    def __idiv__(self, other):
        return self.__itruediv__(other)

    def _comparison_helper(self, other, operation):
        assert_block_structure(self)
        result = self.copy_structure()
        if isinstance(other, MPIBlockVector):
            assert_block_structure(other)
            assert (
                self.nblocks == other.nblocks
            ), 'Number of blocks mismatch: {} != {}'.format(self.nblocks, other.nblocks)
            assert (
                np.array_equal(self._rank_owner, other._rank_owner)
                or self._mpiw.Get_size() == 1
            ), 'MPIBlockVectors must be distributed in same processors'
            assert self._mpiw == other._mpiw, 'Need to have same communicator'

            for i in self._owned_blocks:
                result.set_block(i, operation(self.get_block(i), other.get_block(i)))
            return result
        elif isinstance(other, BlockVector):
            raise RuntimeError('Operation not supported by MPIBlockVector')
        elif isinstance(other, np.ndarray):
            raise RuntimeError('Operation not supported by MPIBlockVector')
        elif np.isscalar(other):
            for i in self._owned_blocks:
                result.set_block(i, operation(self.get_block(i), other))
            return result
        else:
            raise NotImplementedError('Operation not supported by MPIBlockVector')

    def __le__(self, other):
        return self._comparison_helper(other, operator.le)

    def __lt__(self, other):
        return self._comparison_helper(other, operator.lt)

    def __ge__(self, other):
        return self._comparison_helper(other, operator.ge)

    def __gt__(self, other):
        return self._comparison_helper(other, operator.gt)

    def __eq__(self, other):
        return self._comparison_helper(other, operator.eq)

    def __ne__(self, other):
        return self._comparison_helper(other, operator.ne)

    def __contains__(self, item):
        other = item
        assert_block_structure(self)
        if np.isscalar(other):
            contains = False
            for i in self._owned_blocks:
                if other in self.get_block(i):
                    contains = True
            return bool(self._mpiw.allreduce(contains, op=mpi4py.MPI.SUM))
        else:
            raise NotImplementedError('Operation not supported by MPIBlockVector')

    def get_block(self, key):
        owner = self._rank_owner[key]
        rank = self._mpiw.Get_rank()
        assert owner == rank or owner < 0, 'Block {} not own by processor {}'.format(
            key, rank
        )
        return self._block_vector.get_block(key)

    def set_block(self, key, value):
        owner = self._rank_owner[key]
        rank = self._mpiw.Get_rank()
        assert owner == rank or owner < 0, 'Block {} not owned by processor {}'.format(
            key, rank
        )

        self._block_vector.set_block(key, value)
        self._set_block_size(key, value.size)

    def _has_equal_structure(self, other):
        if not (isinstance(other, MPIBlockVector) or isinstance(other, BlockVector)):
            return False
        if self.nblocks != other.nblocks:
            return False
        if isinstance(other, MPIBlockVector):
            if (
                self.owned_blocks != other.owned_blocks
            ).any() and self._mpiw.Get_size() != 1:
                return False
        for ndx in self.owned_blocks:
            block1 = self.get_block(ndx)
            block2 = other.get_block(ndx)
            if isinstance(block1, BlockVector):
                if not isinstance(block2, BlockVector):
                    return False
                if not block1._has_equal_structure(block2):
                    return False
            elif isinstance(block2, BlockVector):
                return False
        return True

    def __getitem__(self, item):
        if not self._has_equal_structure(item):
            raise ValueError(
                'MIPBlockVector.__getitem__ only accepts slices in the form of MPIBlockVectors of the same structure'
            )
        res = self.copy_structure()
        for ndx in self.owned_blocks:
            block = self.get_block(ndx)
            res.set_block(ndx, block[item.get_block(ndx)])

    def __setitem__(self, key, value):
        if not (
            self._has_equal_structure(key)
            and (self._has_equal_structure(value) or np.isscalar(value))
        ):
            raise ValueError(
                'MPIBlockVector.__setitem__ only accepts slices in the form of MPIBlockVectors of the same structure'
            )
        if np.isscalar(value):
            for ndx in self.owned_blocks:
                block = self.get_block(ndx)
                block[key.get_block(ndx)] = value
        else:
            for ndx in self.owned_blocks:
                block = self.get_block(ndx)
                block[key.get_block(ndx)] = value.get_block(ndx)

    def __str__(self):
        msg = '{}{}:\n'.format(self.__class__.__name__, self.bshape)
        for idx in range(self.nblocks):
            msg += '{}: Owned by processor {}\n'.format(idx, self._rank_owner[idx])

        return msg

    def __repr__(self):
        return '{}{}'.format(self.__class__.__name__, self.bshape)

    def pprint(self, root=0):
        """Prints BlockVector in pretty format"""
        assert_block_structure(self)
        msg = self.__repr__() + '\n'
        num_processors = self._mpiw.Get_size()
        local_mask = self._owned_mask.flatten()
        receive_data = np.empty(num_processors * self.nblocks, dtype=bool)
        self._mpiw.Allgather(local_mask, receive_data)
        processor_to_mask = np.split(receive_data, num_processors)

        global_mask = np.zeros(self.nblocks, dtype=bool)

        for bid in range(self.nblocks):
            owner = self._rank_owner[bid]
            if owner >= 0:
                global_mask[bid] = processor_to_mask[owner][bid]
            else:
                # checks only the mask of one of them since all must have the same
                global_mask[bid] = processor_to_mask[0][bid]

            disp_owner = self._rank_owner[bid] if self._rank_owner[bid] >= 0 else 'All'
            is_none = '' if global_mask[bid] else 'None'
            repn = 'Owned by {} Shape({},){}'.format(
                disp_owner, self._brow_lengths[bid], is_none
            )
            msg += '{}: {}\n'.format(bid, repn)
        if self._mpiw.Get_rank() == root:
            print(msg)

    def __len__(self):
        raise NotImplementedError('Use size or nblocks')

    def __iter__(self):
        raise NotImplementedError('Not supported by MPIBlockVector')

    def std(self, axis=None, dtype=None, out=None, ddof=0, keepdims=False):
        raise RuntimeError('Operation not supported by MPIBlockVector')

    def var(self, axis=None, dtype=None, out=None, ddof=0, keepdims=False):
        raise RuntimeError('Operation not supported by MPIBlockVector')

    def cumprod(self, axis=None, dtype=None, out=None):
        raise RuntimeError('Operation not supported by MPIBlockVector')

    def cumsum(self, axis=None, dtype=None, out=None):
        raise RuntimeError('Operation not supported by MPIBlockVector')

    def tolist(self):
        """
        Disable `np.ndarray.tolist` as it is not supported.
        """
        raise RuntimeError('Operation not supported by MPIBlockVector')

    def flatten(self, order='C'):
        raise RuntimeError('Operation not supported by MPIBlockVector')

    def ravel(self, order='C'):
        raise RuntimeError('Operation not supported by MPIBlockVector')



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import (
    numpy as np,
    numpy_available,
    pandas as pd,
    pandas_available,
    scipy_available,
)

from pyomo.common.errors import DeveloperError
import pyomo.common.unittest as unittest

if not (numpy_available and scipy_available):
    raise unittest.SkipTest("Pyomo.DoE needs scipy and numpy to run tests")

if scipy_available:
    from pyomo.contrib.doe import DesignOfExperiments
    from pyomo.contrib.doe.tests.experiment_class_example_flags import (
        BadExperiment,
        RooneyBieglerExperimentFlag,
    )
    from pyomo.contrib.parmest.examples.rooney_biegler.rooney_biegler import (
        RooneyBieglerExperiment,
    )

from pyomo.contrib.doe.examples.rooney_biegler_doe_example import run_rooney_biegler_doe
from pyomo.opt import SolverFactory

ipopt_available = SolverFactory("ipopt").available()


def get_rooney_biegler_experiment_flag():
    """Get a fresh RooneyBieglerExperimentFlag instance for testing.

    Creates a new experiment instance that supports the flag parameter
    for creating incomplete models for error testing.
    """
    data = pd.DataFrame(data=[[5, 15.6]], columns=['hour', 'y'])
    data_point = data.iloc[0]

    return RooneyBieglerExperimentFlag(
        data=data_point,
        theta={'asymptote': 15, 'rate_constant': 0.5},
        measure_error=0.1,
    )


def get_standard_args(experiment, fd_method, obj_used, flag):
    args = {}
    args['experiment'] = experiment
    args['fd_formula'] = fd_method
    args['step'] = 1e-3
    args['objective_option'] = obj_used
    args['scale_constant_value'] = 1
    args['scale_nominal_param_value'] = True
    args['prior_FIM'] = None
    args['jac_initial'] = None
    args['fim_initial'] = None
    args['L_diagonal_lower_bound'] = 1e-7
    # Make solver object with
    # good linear subroutines
    solver = SolverFactory("ipopt")
    solver.options["linear_solver"] = "ma57"
    solver.options["halt_on_ampl_error"] = "yes"
    solver.options["max_iter"] = 3000
    args['solver'] = solver
    args['tee'] = False
    if flag is not None:
        args['get_labeled_model_args'] = {"flag": flag}
    args['_Cholesky_option'] = True
    args['_only_compute_fim_lower'] = True
    return args


@unittest.skipIf(not numpy_available, "Numpy is not available")
@unittest.skipIf(not scipy_available, "scipy is not available")
@unittest.skipIf(not pandas_available, "pandas is not available")
class TestDoEErrors(unittest.TestCase):
    def test_experiment_none_error(self):
        fd_method = "central"
        obj_used = "pseudo_trace"
        flag_val = 1  # Value for faulty model build mode - 1: No exp outputs

        with self.assertRaisesRegex(
            ValueError, "Experiment object must be provided to perform DoE."
        ):
            # Experiment provided as None
            DoE_args = get_standard_args(None, fd_method, obj_used, flag_val)

            doe_obj = DesignOfExperiments(**DoE_args)

    def test_reactor_check_no_get_labeled_model(self):
        fd_method = "central"
        obj_used = "pseudo_trace"
        flag_val = 1  # Value for faulty model build mode - 1: No exp outputs

        experiment = BadExperiment()

        with self.assertRaisesRegex(
            ValueError,
            "The experiment object must have a ``get_labeled_model`` function",
        ):
            DoE_args = get_standard_args(experiment, fd_method, obj_used, flag_val)

            doe_obj = DesignOfExperiments(**DoE_args)

    def test_reactor_check_no_experiment_outputs(self):
        fd_method = "central"
        obj_used = "pseudo_trace"
        flag_val = 1  # Value for faulty model build mode - 1: No exp outputs

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag_val)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            RuntimeError,
            "Experiment model does not have suffix " + '"experiment_outputs".',
        ):
            doe_obj.create_doe_model()

    def test_reactor_check_no_measurement_error(self):
        fd_method = "central"
        obj_used = "pseudo_trace"
        flag_val = 2  # Value for faulty model build mode - 2: No meas error

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag_val)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            RuntimeError,
            "Experiment model does not have suffix " + '"measurement_error".',
        ):
            doe_obj.create_doe_model()

    def test_reactor_check_no_experiment_inputs(self):
        fd_method = "central"
        obj_used = "pseudo_trace"
        flag_val = 3  # Value for faulty model build mode - 3: No exp inputs/design vars

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag_val)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            RuntimeError,
            "Experiment model does not have suffix " + '"experiment_inputs".',
        ):
            doe_obj.create_doe_model()

    def test_reactor_check_no_unknown_parameters(self):
        fd_method = "central"
        obj_used = "pseudo_trace"
        flag_val = 4  # Value for faulty model build mode - 4: No unknown params

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag_val)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            RuntimeError,
            "Experiment model does not have suffix " + '"unknown_parameters".',
        ):
            doe_obj.create_doe_model()

    def test_reactor_check_bad_prior_size(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        prior_FIM = np.ones((5, 5))

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)
        DoE_args['prior_FIM'] = prior_FIM

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            ValueError,
            "Shape of FIM provided should be n parameters by n parameters, "
            "or {} by {}, FIM provided has shape {} by {}".format(
                2, 2, prior_FIM.shape[0], prior_FIM.shape[1]
            ),
        ):
            doe_obj.create_doe_model()

    def test_reactor_check_bad_prior_negative_eigenvalue(self):
        from pyomo.contrib.doe.doe import _SMALL_TOLERANCE_DEFINITENESS

        fd_method = "central"
        obj_used = "pseudo_trace"

        prior_FIM = -np.ones((2, 2))

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)
        DoE_args['prior_FIM'] = prior_FIM

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            ValueError,
            r"FIM provided is not positive definite. It has one or more "
            r"negative eigenvalue\(s\) less than -{:.1e}".format(
                _SMALL_TOLERANCE_DEFINITENESS
            ),
        ):
            doe_obj.create_doe_model()

    def test_reactor_check_bad_prior_not_symmetric(self):
        from pyomo.contrib.doe.utils import _SMALL_TOLERANCE_SYMMETRY

        fd_method = "central"
        obj_used = "pseudo_trace"

        prior_FIM = np.zeros((2, 2))
        prior_FIM[0, 1] = 1e-3

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)
        DoE_args['prior_FIM'] = prior_FIM

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            ValueError,
            "FIM provided is not symmetric using absolute tolerance {}".format(
                _SMALL_TOLERANCE_SYMMETRY
            ),
        ):
            doe_obj.create_doe_model()

    def test_reactor_check_bad_jacobian_init_size(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        jac_init = np.ones((5, 5))

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)
        DoE_args['jac_initial'] = jac_init

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            ValueError,
            "Shape of Jacobian provided should be n experiment outputs "
            "by n parameters, or {} by {}, Jacobian provided has "
            "shape {} by {}".format(1, 2, jac_init.shape[0], jac_init.shape[1]),
        ):
            doe_obj.create_doe_model()

    def test_reactor_check_unbuilt_update_FIM(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        FIM_update = np.ones((2, 2))

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            RuntimeError,
            "``fim`` is not defined on the model provided. "
            "Please build the model first.",
        ):
            doe_obj.update_FIM_prior(FIM=FIM_update)

    def test_reactor_check_none_update_FIM(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        FIM_update = None

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            ValueError,
            "FIM input for update_FIM_prior must be a 2D, square numpy array.",
        ):
            doe_obj.update_FIM_prior(FIM=FIM_update)

    def test_reactor_check_results_file_name(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            ValueError, "``results_file`` must be either a Path object or a string."
        ):
            doe_obj.run_doe(results_file=int(15))

    def test_reactor_check_measurement_and_output_length_match(self):
        fd_method = "central"
        obj_used = "pseudo_trace"
        flag_val = (
            5  # Value for faulty model build mode - 5: Mismatch error and output length
        )

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag_val)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            ValueError,
            "Number of experiment outputs, {}, and length of measurement error, "
            "{}, do not match. Please check model labeling.".format(1, 2),
        ):
            doe_obj.create_doe_model()

    @unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
    def test_reactor_grid_search_des_range_inputs(self):
        fd_method = "central"
        obj_used = "determinant"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        design_ranges = {"not": [1, 5, 3], "correct": [300, 700, 3]}

        with self.assertRaisesRegex(
            ValueError,
            "Design ranges keys must be a subset of experimental design names.",
        ):
            doe_obj.compute_FIM_full_factorial(
                design_ranges=design_ranges, method="sequential"
            )

    @unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
    def test_reactor_premature_figure_drawing(self):
        fd_method = "central"
        obj_used = "determinant"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            RuntimeError,
            "Results must be provided "
            "or the compute_FIM_full_factorial function must be run.",
        ):
            doe_obj.draw_factorial_figure()

    @unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
    def test_reactor_figure_drawing_no_des_var_names(self):
        fd_method = "central"
        obj_used = "determinant"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        design_ranges = {"hour": [1, 7, 2]}

        doe_obj.compute_FIM_full_factorial(
            design_ranges=design_ranges, method="sequential"
        )

        with self.assertRaisesRegex(
            ValueError,
            "If results object is provided, you must "
            "include all the design variable names.",
        ):
            doe_obj.draw_factorial_figure(results=doe_obj.fim_factorial_results)

    @unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
    def test_reactor_figure_drawing_no_sens_names(self):
        fd_method = "central"
        obj_used = "determinant"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        design_ranges = {"hour": [1, 7, 2]}

        doe_obj.compute_FIM_full_factorial(
            design_ranges=design_ranges, method="sequential"
        )

        with self.assertRaisesRegex(
            ValueError, "``sensitivity_design_variables`` must be included."
        ):
            doe_obj.draw_factorial_figure()

    @unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
    def test_reactor_figure_drawing_no_fixed_names(self):
        fd_method = "central"
        obj_used = "determinant"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        design_ranges = {"hour": [1, 7, 2]}

        doe_obj.compute_FIM_full_factorial(
            design_ranges=design_ranges, method="sequential"
        )

        with self.assertRaisesRegex(
            ValueError, "``fixed_design_variables`` must be included."
        ):
            doe_obj.draw_factorial_figure(sensitivity_design_variables={"dummy": "var"})

    @unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
    def test_reactor_figure_drawing_bad_fixed_names(self):
        fd_method = "central"
        obj_used = "determinant"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        design_ranges = {"hour": [1, 7, 2]}

        doe_obj.compute_FIM_full_factorial(
            design_ranges=design_ranges, method="sequential"
        )

        with self.assertRaisesRegex(
            ValueError,
            "Fixed design variables do not all appear in the results object keys.",
        ):
            doe_obj.draw_factorial_figure(
                sensitivity_design_variables={"hour": 1},
                fixed_design_variables={"bad": "entry"},
            )

    @unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
    def test_reactor_figure_drawing_bad_sens_names(self):
        fd_method = "central"
        obj_used = "determinant"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        design_ranges = {"hour": [1, 7, 2]}

        doe_obj.compute_FIM_full_factorial(
            design_ranges=design_ranges, method="sequential"
        )

        with self.assertRaisesRegex(
            ValueError,
            "Sensitivity design variables do not all appear "
            "in the results object keys.",
        ):
            doe_obj.draw_factorial_figure(
                sensitivity_design_variables={"bad": "entry"},
                fixed_design_variables={"hour": 1},
            )

    def test_reactor_check_get_FIM_without_FIM(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            RuntimeError,
            "Model provided does not have variable `fim`. Please make sure "
            "the model is built properly before calling `get_FIM`",
        ):
            doe_obj.get_FIM()

    def test_reactor_check_get_sens_mat_without_model(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            RuntimeError,
            "Model provided does not have variable `sensitivity_jacobian`. "
            "Please make sure the model is built properly before calling "
            "`get_sensitivity_matrix`",
        ):
            doe_obj.get_sensitivity_matrix()

    def test_reactor_check_get_exp_inputs_without_model(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            RuntimeError,
            "Model provided does not have expected structure. "
            "Please make sure model is built properly before "
            "calling `get_experiment_input_values`",
        ):
            doe_obj.get_experiment_input_values()

    def test_reactor_check_get_exp_outputs_without_model(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            RuntimeError,
            "Model provided does not have expected structure. Please make "
            "sure model is built properly before calling "
            "`get_experiment_output_values`",
        ):
            doe_obj.get_experiment_output_values()

    def test_reactor_check_get_unknown_params_without_model(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            RuntimeError,
            "Model provided does not have expected structure. Please make "
            "sure model is built properly before calling "
            "`get_unknown_parameter_values`",
        ):
            doe_obj.get_unknown_parameter_values()

    def test_reactor_check_get_meas_error_without_model(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            RuntimeError,
            "Model provided does not have expected structure. Please make "
            "sure model is built properly before calling "
            "`get_measurement_error_values`",
        ):
            doe_obj.get_measurement_error_values()

    def test_multiple_exp_not_implemented_seq(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            NotImplementedError, "Multiple experiment optimization not yet supported."
        ):
            doe_obj.run_multi_doe_sequential(N_exp=1)

    def test_multiple_exp_not_implemented_sim(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            NotImplementedError, "Multiple experiment optimization not yet supported."
        ):
            doe_obj.run_multi_doe_simultaneous(N_exp=1)

    def test_update_unknown_parameter_values_not_implemented_seq(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            NotImplementedError, "Updating unknown parameter values not yet supported."
        ):
            doe_obj.update_unknown_parameter_values()

    @unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
    def test_bad_FD_generate_scens(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            DeveloperError,
            "Internal Pyomo implementation error:\n"
            "        "
            "'Finite difference option not recognized. Please contact the\n"
            "        "
            "developers as you should not see this error.'\n"
            "    "
            "Please report this to the Pyomo Developers.",
        ):
            doe_obj.fd_formula = "bad things"
            doe_obj._generate_scenario_blocks()

    @unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
    def test_bad_FD_seq_compute_FIM(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            DeveloperError,
            "Internal Pyomo implementation error:\n"
            "        "
            "'Finite difference option not recognized. Please contact the\n"
            "        "
            "developers as you should not see this error.'\n"
            "    "
            "Please report this to the Pyomo Developers.",
        ):
            doe_obj.fd_formula = "bad things"
            doe_obj.compute_FIM(method="sequential")

    def test_bad_objective(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            DeveloperError,
            "Internal Pyomo implementation error:\n"
            "        "
            "'Objective option not recognized. Please contact the developers as\n"
            "        "
            "you should not see this error.'\n"
            "    "
            "Please report this to the Pyomo Developers.",
        ):
            doe_obj.objective_option = "bad things"
            doe_obj.create_objective_function()

    def test_no_model_for_objective(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            RuntimeError,
            "Model provided does not have variable `fim`. Please make "
            "sure the model is built properly before creating the objective.",
        ):
            doe_obj.create_objective_function()

    @unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
    def test_bad_compute_FIM_option(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        experiment = get_rooney_biegler_experiment_flag()

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            ValueError,
            "The method provided, {}, must be either `sequential` or `kaug`".format(
                "Bad Method"
            ),
        ):
            doe_obj.compute_FIM(method="Bad Method")

    @unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
    def test_invalid_trace_without_cholesky(self):
        fd_method = "central"
        obj_used = "trace"

        experiment = run_rooney_biegler_doe()["experiment"]

        DoE_args = get_standard_args(experiment, fd_method, obj_used, flag=None)
        DoE_args['_Cholesky_option'] = False

        doe_obj = DesignOfExperiments(**DoE_args)
        doe_obj.create_doe_model()

        with self.assertRaisesRegex(
            ValueError,
            "objective_option='trace' currently only implemented with ``_Cholesky option=True``.",
        ):
            doe_obj.create_objective_function()


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.environ as pyo

from pyomo.contrib.iis.mis import compute_infeasibility_explanation


class TestMIS(unittest.TestCase):
    def test_trivial_quad(self):
        m = pyo.ConcreteModel("Trivial Quad")
        m.x = pyo.Var([1, 2], bounds=(0, 1))
        m.y = pyo.Var(bounds=(0, 1))
        m.c = pyo.Constraint(expr=m.x[1] * m.x[2] == -1)
        m.d = pyo.Constraint(expr=m.x[1] + m.y >= 1)
        # Note: this particular little problem is quadratic
        # As of 18Feb2024 DLW is not sure the explanation code works
        # with solvers other than ipopt
        ipopt = pyo.SolverFactory("ipopt")
        compute_infeasibility_explanation(m, solver=ipopt)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Utility functions for the high confidence stopping rule.

This stopping criterion operates by estimating the amount of missing optima,
and stops once the estimated mass of missing optima is within an acceptable
range, given some confidence.

"""

from collections import Counter
from math import log, sqrt


def num_one_occurrences(observed_obj_vals, tolerance):
    """
    Determines the number of optima that have only been observed once.
    Needed to estimate missing mass of optima.
    """
    obj_value_distribution = Counter(observed_obj_vals)
    sorted_histogram = list(sorted(obj_value_distribution.items()))
    if tolerance == 0:
        return sum(1 for _, count in sorted_histogram if count == 1)
    else:
        # Need to apply a tolerance to each value
        num_obj_vals_only_observed_once = 0
        for i, tup in enumerate(sorted_histogram):
            obj_val, count = tup
            if count == 1:
                # look at previous and next elements to make sure that they are
                # not within the tolerance
                if i > 0 and obj_val - sorted_histogram[i - 1][0] <= tolerance:
                    continue
                if (
                    i < len(sorted_histogram) - 1
                    and sorted_histogram[i + 1][0] - obj_val <= tolerance
                ):
                    continue
                num_obj_vals_only_observed_once += 1
        return num_obj_vals_only_observed_once


def should_stop(solutions, stopping_mass, stopping_delta, tolerance):
    """
    Determines if the missing mass of unseen local optima is acceptable
    based on the High Confidence stopping rule.
    """
    f = num_one_occurrences(solutions, tolerance)
    n = len(solutions)
    if n == 0:
        return False  # Do not stop if no solutions have been found.
    d = stopping_delta
    c = stopping_mass
    confidence = f / n + (2 * sqrt(2) + sqrt(3)) * sqrt(log(3 / d) / n)
    return confidence < c
